#pragma warning( disable : 4786)

#include "config_win32.h"
#include "config_unix.h"
#include "debug.h"

#include <map>

#include "profiler.h"



//#define USE_SLOG

#ifdef USE_SLOG
extern "C" {
#include "slog.h"
}
#else
struct SLOG_STREAM;
#endif

/* Perf Counter stuff */
#define NCOUNTER N_STATES
static LARGE_INTEGER accum[NCOUNTER], currentCounter[NCOUNTER];
static long count[NCOUNTER];

#ifdef USE_SLOG
static SLOG_STREAM *slogStream;
#endif

static double freq;
static LARGE_INTEGER offset;

static unsigned long currentSource;
static int currentSourceAttributed;

/*
 * Convert to seconds since the start of the run.
 */
static inline double convertCountToTime(LARGE_INTEGER c)
{
	return (double) (c.QuadPart - offset.QuadPart) / freq;
}

static inline double convertIntervalToTime(LARGE_INTEGER c)
{
	return (double) c.QuadPart / freq;
}

#define MAX_BLIT_ENTRIES 500000
struct BlitEntry
{
    int w;
    int h;
    LARGE_INTEGER elap;
};
BlitEntry blitEntries[MAX_BLIT_ENTRIES];
static int nBlitEntries = 0;

static void initBlitEntries()
{
    nBlitEntries = 0;
}


void addBlitEntry(int w, int h, LARGE_INTEGER *elap)
{
    if (nBlitEntries < MAX_BLIT_ENTRIES)
    {
	BlitEntry *be = &(blitEntries[nBlitEntries++]);
	be->w = w;
	be->h = h;
	be->elap = *elap;
    }
}

static void dumpBlitEntries()
{
    FILE *fp = fopen("blitEntries.out", "w");
    if (fp == 0)
	return;
    for (int i = 0; i < nBlitEntries; i++)
    {
	BlitEntry *be = &(blitEntries[i]);
	   
	fprintf(fp, "%d %d %lf\n", be->w, be->h, convertIntervalToTime(be->elap));
    }
    fclose(fp);
}


/* Stuff for storing the log records */

class SsrcMap
{
private:
	typedef std::map<unsigned long, int, std::less<unsigned long> > maptype;
	maptype map_;
	int nextPid_;

public:
	SsrcMap() {
		nextPid_ = 1;
	}
	int findPid(unsigned long ssrc) {
		int thisPid;

		maptype::iterator it;
		it = map_.find(ssrc);
		if (it == map_.end())
		{
			thisPid = nextPid_++;
			debug_msg("New ssrc found, assigning pid %d\n", thisPid);
			map_[ssrc] = thisPid;
		}
		else
		{
			thisPid = (*it).second;
		}
		return thisPid;
	}
};

#if 1

#define statedef(id, name, color) color,
char *state_colors[] = {
#include "statedef.h"
};

#undef statedef
#define statedef(id, name, color) name,
char *state_names[] = {
#include "statedef.h"
};

#define state_count (sizeof(state_names) / sizeof(state_names[0]))

#else
char *state_names[] = {""};
#define state_count 0
#endif


static SsrcMap ssrcMap;

#ifdef USE_SLOG

struct LogRecord {
	LARGE_INTEGER start;
	LARGE_INTEGER elap;
	int eventType;
	int isSsrcRecord;
	unsigned long data;
};


class LogRecordSet {
private:
	LogRecord *records_;
	int nRecordsAllocated_;
	int nRecords_;
	SLOG_STREAM *slog_;

public:
	LogRecordSet(int size) {
		records_ = new LogRecord[size];
		nRecordsAllocated_ = size;
		nRecords_ = 0;
		slog_ = 0;
	}

	~LogRecordSet() {
		delete records_;
	}

	void setSlog(SLOG_STREAM *s) { slog_ = s; }

	void logSsrcEvent(const LARGE_INTEGER &start, const LARGE_INTEGER &elap,
					  int eventType, unsigned long ssrc)
	{
		if (nRecords_ >= nRecordsAllocated_)
			return;

		LogRecord *r = &(records_[nRecords_]);
		nRecords_++;
		r->start = start;
		r->elap = elap;
		r->eventType = eventType;
		r->isSsrcRecord = 1;
		r->data = ssrc;

		if (nRecords_ == nRecordsAllocated_)
			dumpRecords();
	}

	void logEvent(const LARGE_INTEGER &start, const LARGE_INTEGER &elap,
					  int eventType, unsigned long ssrc)
	{
		if (nRecords_ >= nRecordsAllocated_)
			return;

		LogRecord *r = &(records_[nRecords_]);
		nRecords_++;
		r->start = start;
		r->elap = elap;
		r->eventType = eventType;
		r->isSsrcRecord = 0;
		r->data = ssrc;

		if (nRecords_ == nRecordsAllocated_)
			dumpRecords();
	}

	void dumpRecords()
	{
		writeToSLOG(slog_);
		nRecords_ = 0;
	}

	void writeToSLOG(SLOG_STREAM *slog)
	{
		debug_msg("writing %d entries to slog\n", nRecords_);

		SLOG_Irec irec = SLOG_Irec_Create();

		for (int i = 0; i < nRecords_; i++)
			writeRecordToSLOG(slog, irec, &(records_[i]));

		SLOG_Irec_Free(irec);
	}

	void writeRecordToSLOG(SLOG_STREAM *slog, SLOG_Irec irec, LogRecord *r)
	{
		SLOG_Irec_ReserveSpace( slog, SLOG_RECTYPE_STATIC_DIAG, r->eventType, 1, 1,
			convertCountToTime(r->start));

		// Determine the process id
		
		int pid;
		if (r->isSsrcRecord)
		{
			pid = ssrcMap.findPid(r->data);
		}
		else
		{
			pid = r->data;
		}


	    SLOG_Irec_SetMinRec( irec, SLOG_RECTYPE_STATIC_DIAG, r->eventType, 1, 1, 
				convertCountToTime(r->start), convertIntervalToTime(r->elap), 
				 0, pid,0, 0);
	
	    SLOG_Irec_ToOutputStream( slog, irec );
	}
};

static LogRecordSet logRecordSet(1000000);

#endif

#ifdef debug_msg
#define printf debug_msg
#endif

int stateCount() 
{ 
	return state_count;
}

void initSlog()
{
#ifdef USE_SLOG
	slogStream = SLOG_OpenOutputStream("slog.out");
	logRecordSet.setSlog(slogStream);

    /*
        Tell the API the properties of the slog file by using SLOG_SetXXXX()
    */

    /*  Set the maimum number of frames per directory in slog file  */
    SLOG_SetMaxNumOfFramesPerDir( slogStream, 100 );

    /*  Set the byte size per frame  */
    SLOG_SetFrameByteSize( slogStream, 1000000 );

    /*  Set the reserved space in each frame  */
    /*  SLOG_SetFrameReservedSpace( slog, SLOG_typesz[ min_IntvlRec ] );  */
    /*  Here SLOG_Irec_ReserveSpace() will be used, so set 0 reserve space  */
    SLOG_SetFrameReservedSpace( slogStream, 0 );

    /*  Set the flag for the time order arrangement of the interval record  */
    /*  SLOG_SetIncreasingStarttimeOrder( slog );  */
    SLOG_SetIncreasingEndtimeOrder( slogStream ); 

    /*
        Set the SLOG Preview to be read from a given file, filename,
        and then init SLOG's Preview section
    */ 
    /*  SLOG_SetPreviewName( slog, SLOG_PREVIEWNAME );  */
     SLOG_SetPreviewName( slogStream, "nul:" );

    /*
        This is an optional call to set the Preview Statistics's array
        size, if this function is NOT called, the default size
        SLOG_PSTAT_STATSIZE_MAX will be used.
    */
    SLOG_SetPreviewStatisticsSize( slogStream, 128 );

	   /*  Initialize the SLOG Thread Table  */
    if ( SLOG_TTAB_Open( slogStream ) != SLOG_SUCCESS ) {
        debug_msg(__FILE__":Main() - SLOG_TTAB_Open() fails!\n" );
        exit( 1 );
    }

	int ierr;

    /*  Add the entries of SLOG Thread Table  */
    
    ierr = SLOG_TTAB_AddThreadInfo( slogStream, 0, 0, 0, 0, 0);
    if ( ierr != SLOG_SUCCESS ) {
        debug_msg( __FILE__"Main() - "
                          "SLOG_TTAB_AddTheadInfo() fails!\n" ); 
        /*  exit( 1 );  */
    }

    /*  Finalize the Record Definition Table in SLOG file  */
    ierr = SLOG_TTAB_Close( slogStream );
    if ( ierr != SLOG_SUCCESS ) {
        debug_msg( __FILE__"Main() - "
                          "SLOG_TTAB_Close() fails!\n" ); 
        exit( 1 );
    } 
    
    /*  Set the SLOG Display Profile with a given filename and init SLOG  */
    /*  SLOG_SetProfileName( slog, SLOG_PROFILENAME );  */

    /*  Initialize the SLOG Display Profile Table  */
    if ( SLOG_PROF_Open( slogStream ) != SLOG_SUCCESS ) {
        debug_msg( __FILE__":Main() - SLOG_PROF_Open() fails\n " );
		exit( 1 );
    }

	for (int i = 0; i < state_count; i++)
	{
		/*  Create the Indexes for Message Record's display profile  */
		ierr = SLOG_PROF_AddIntvlInfo( slogStream, i,
									   "State", state_names[i], state_colors[i]);
		printf("add int type %d %s %s\n", i, state_names[i], state_colors[i]);
		if ( ierr != SLOG_SUCCESS ) {
			debug_msg( __FILE__":Main() - "
							  "SLOG_PROF_AddIntvlInfo( Message ) fails!\n" );
			exit( 1 );
		}
	}

    /*  Finalize the Display Profile Table in SLOG file  */
    ierr = SLOG_PROF_Close( slogStream );
    if ( ierr != SLOG_SUCCESS ) {
        debug_msg( __FILE__":Main() - "
                          "SLOG_PROF_Close() fails!\n" );
        exit( 1 );
    }
   /*  Initialize the SLOG Record Definition Table  */
    if ( SLOG_RDEF_Open( slogStream ) != SLOG_SUCCESS ) {
        debug_msg( __FILE__":Main() - SLOG_RDEF_Open() fails\n " );
        exit( 1 );
    }

	for (i = 0; i < state_count; i++)
	{
		/*  Create the Indexes for Message Record's Definition  */
		ierr = SLOG_RDEF_AddRecDef( slogStream, i,
									1, 1,
									0, 0 );
		if ( ierr != SLOG_SUCCESS ) {
			debug_msg(  __FILE__":Main() - "
							  "SLOG_RDEF_AddRecDef( Message ) fails!\n" ); 
			exit( 1 );
		}
	}	/*  Finalize the Record Definition Table in SLOG file  */
    ierr = SLOG_RDEF_Close( slogStream );
    if ( ierr != SLOG_SUCCESS ) {
        debug_msg(  __FILE__":Main() - "
                          "SLOG_RDEF_Close() fails!\n" );
        exit( 1 );
    } 
#endif
}

void initCounters()
{
	for (int i = 0; i < NCOUNTER; i++)
	{
		count[i] = 0;
		accum[i].QuadPart = 0;
	}
	LARGE_INTEGER freql;
    QueryPerformanceFrequency(&freql);
	freq = (double) freql.QuadPart;
	QueryPerformanceCounter(&offset);
	initBlitEntries();
#ifdef USE_SLOG
	initSlog();
#endif

	currentSourceAttributed = 0;
}

void profilerSourceAttribute(unsigned long src)
{
// debug_msg("attribute to %d\n", src);
	currentSourceAttributed = 1;
	currentSource = src;
}

void profilerSourceDeattribute()
{
	currentSourceAttributed = 0;
	currentSource = 0;
}

void startCounter(int which)
{
	QueryPerformanceCounter(&(currentCounter[which]));
	count[which]++;
}

void stopCounterC(int which)
{
    stopCounter(which);
}
    
void stopCounter(int which, LARGE_INTEGER *elapPtr)
{
	LARGE_INTEGER now, elap;
	const LARGE_INTEGER &start = currentCounter[which];

	QueryPerformanceCounter(&now);

	elap.QuadPart = now.QuadPart - currentCounter[which].QuadPart;
	accum[which].QuadPart += elap.QuadPart;

	if (elapPtr != 0)
	    *elapPtr = elap;

	if (which == STATE_RUNTIME)
		return;

#ifdef USE_SLOG
	if (currentSourceAttributed)
		logRecordSet.logSsrcEvent(start, elap, which, currentSource);
	else
		logRecordSet.logEvent(start, elap, which, currentSource);
#endif

}

void dumpCounters()
{
    FILE *fp =fopen("counters.out", "w");
    LARGE_INTEGER freq;
    QueryPerformanceFrequency(&freq);
    fprintf(fp, "counter frequency=%Lu\n", freq.QuadPart);
	fprintf(fp, "counter\ttime\t\tcount\tpct total\twhat\n");
    for (int i = 0; i< NCOUNTER; i++)
    {
		char *name;
		if (i < state_count && state_names[i] != 0)
			name = state_names[i];
		else
			name = "";


		fprintf(fp, "%d\t%lf\t%d\t%6.2f\t%s\n", i, 
	    (double) accum[i].QuadPart /(double) freq.QuadPart, 
	    count[i],
	    100.0 * (double) accum[i].QuadPart /(double) accum[0].QuadPart ,
		name
	    );
    }
    fclose(fp);
    dumpBlitEntries();

#ifdef USE_SLOG
	logRecordSet.dumpRecords();
	SLOG_CloseOutputStream( slogStream);
#endif
}
