/*
 * x509name.c
 *
 * Copyright (C) AB Strakt 2001, All rights reserved
 *
 * X.509 Name handling, mostly thin wrapping.
 * See the file RATIONALE for a short explanation of why this module was written.
 *
 * Reviewed 2001-07-23
 */
#include <Python.h>
#define crypto_MODULE
#include "crypto.h"

static char *CVSid = "@(#) $Id: x509name.c,v 1.8 2004/01/26 23:17:31 olson Exp $";


/*
 * Constructor for X509Name, never called by Python code directly
 *
 * Arguments: name    - A "real" X509_NAME object
 *            dealloc - Boolean value to specify whether the destructor should
 *                      free the "real" X509_NAME object
 * Returns:   The newly created X509Name object
 */
crypto_X509NameObj *
crypto_X509Name_New(X509_NAME *name, int dealloc)
{
    crypto_X509NameObj *self;

    self = PyObject_New(crypto_X509NameObj, &crypto_X509Name_Type);

    if (self == NULL)
        return NULL;

    self->x509_name = name;
    self->dealloc = dealloc;
    self->parent_cert = NULL;

    return self;
}

/*
 * Deallocate the memory used by the X509Name object
 *
 * Arguments: self - The X509Name object
 * Returns:   None
 */
static void
crypto_X509Name_dealloc(crypto_X509NameObj *self)
{
    /* Sometimes we don't have to dealloc this */
    if (self->dealloc)
        X509_NAME_free(self->x509_name);

    Py_XDECREF(self->parent_cert);

    PyObject_Del(self);
}

/*
 * Return a name string given a X509_NAME object and a name identifier. Used
 * by the getattr function.
 *
 * Arguments: name - The X509_NAME object
 *            nid  - The name identifier
 * Returns:   The name as a Python string object
 */
static PyObject *
get_name_by_nid(X509_NAME *name, int nid)
{
    void *buf;
    int len, xlen;
    PyObject *str;

    if ((len = X509_NAME_get_text_by_NID(name, nid, NULL, 0)) == -1)
    {
        Py_INCREF(Py_None);
        return Py_None;
    }

    len++;
    buf = PyMem_Malloc(len);
    if (buf == NULL)
    {
	printf("malloc %d failed\n", len);
        return NULL;
    }

    xlen = X509_NAME_get_text_by_NID(name, nid, buf, len);
    str = PyString_FromStringAndSize(buf, xlen);
    PyMem_Free(buf);

    return str;
}

/*
 * Given a X509_NAME object and a name identifier, set the corresponding
 * attribute to the given string. Used by the setattr function.
 *
 * Arguments: name  - The X509_NAME object
 *            nid   - The name identifier
 *            value - The string to set
 * Returns:   0 for success, -1 on failure
 */
static int
set_name_by_nid(X509_NAME *name, int nid, PyObject *value)
{
    X509_NAME_ENTRY *ne;
    int i, entry_count, temp_nid;

    /* If there's an old entry for this NID, remove it */
    entry_count = X509_NAME_entry_count(name);
    for (i = 0; i < entry_count; i++)
    {
        ne = X509_NAME_get_entry(name, i);
        temp_nid = OBJ_obj2nid(X509_NAME_ENTRY_get_object(ne));
        if (temp_nid == nid)
        {
            ne = X509_NAME_delete_entry(name, i);
            X509_NAME_ENTRY_free(ne);
            break;
        }
    }

    /* Add the new entry */
    if (!X509_NAME_add_entry_by_NID(name, nid, MBSTRING_ASC,
            PyString_AsString(value), -1, -1, 0))
    {
        exception_from_error_queue();
        return -1;
    }
    return 0;
}

static char crypto_X509Name_add_doc[] = "\n\
Add an element to the name\n\
\n\
Arguments: self - The X509 object\n\
           args - The Python argument tuple, should be:\n\
             name - the attribute name\n\
             value - the value to add\n\
Returns:   None\n\
";

/*
 * Add an element to the name.
 *
 * Arguments: self  - The X509Name object
 *            name  - The attribute name
 *            value - The value to set
 */
static PyObject *
crypto_X509Name_add(crypto_X509NameObj *self, PyObject *args)
{
    char *name, *val;
    int nid;
    
    if (!PyArg_ParseTuple(args, "ss:add", &name, &val))
	return NULL;

    if ((nid = OBJ_txt2nid(name)) == NID_undef)
    {
        PyErr_SetString(PyExc_AttributeError, "No such attribute");
        return NULL;
    }
    
    /* Add the new entry */
    if (!X509_NAME_add_entry_by_NID(self->x509_name, nid, MBSTRING_ASC,
				    val, -1, -1, 1))
    {
        PyErr_SetString(PyExc_AttributeError, "Add failed");
	return NULL;
    }

    Py_INCREF(Py_None);
    return Py_None;
}

static char crypto_X509Name_get_hash_doc[] = "\n\
Return the has value of this name\n\
\n\
Arguments: self - The X509 object\n\
           args - The Python argument tuple, should be empty\n\
Returns:   None\n\
";

static PyObject *
crypto_X509Name_get_hash(crypto_X509NameObj *self, PyObject *args)
{
    char hstr[16];
    unsigned long h;
    
    if (!PyArg_ParseTuple(args, ":get_hash"))
	return NULL;

    h = X509_NAME_hash(self->x509_name);
    snprintf(hstr, sizeof(hstr), "%08lx", h);
    return PyString_FromString(hstr);
}

static char crypto_X509Name_get_der_doc[] = "\n\
Return the DER encodeing of this name\n\
\n\
Arguments: self - The X509 object\n\
           args - The Python argument tuple, should be empty\n\
Returns:   None\n\
";

static PyObject *
crypto_X509Name_get_der(crypto_X509NameObj *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, ":get_der"))
	return NULL;

    i2d_X509_NAME(self->x509_name, 0);
    return PyString_FromStringAndSize(self->x509_name->bytes->data,
				      self->x509_name->bytes->length);
				      
}

static char crypto_X509Name_get_name_components_doc[] = "\n\
Returns the split-up components of this name.\n\
\n\
Arguments: self - The X509 object\n\
           args - The Python argument tuple, should be empty\n\
Returns:   List of tuples (name, value).\n\
";

static PyObject *
crypto_X509Name_get_name_components(crypto_X509NameObj *self, PyObject *args)
{
    int n, i;
    X509_NAME *name = self->x509_name;
    PyObject *list;
    
    if (!PyArg_ParseTuple(args, ":get_name_components"))
	return NULL;

    n = X509_NAME_entry_count(name);
    list = PyList_New(n);
    for (i = 0; i < n; i++)
    {
	X509_NAME_ENTRY *ent;
	ASN1_OBJECT *fname;
	ASN1_STRING *fval;
	int nid;
	int l;
	unsigned char *str;
	PyObject *tuple;
	const char *sn;
	char tmp_buf[80];

	ent = X509_NAME_get_entry(name, i);

	fname = X509_NAME_ENTRY_get_object(ent);
	fval = X509_NAME_ENTRY_get_data(ent);

	l = ASN1_STRING_length(fval);
	str = ASN1_STRING_data(fval);

	nid = OBJ_obj2nid(fname);


	/*
	 * If we don't have an NID for the name, or if it doesn't have
	 * a shortname, return the OID of the name.
	 */

	if ((nid == NID_undef) || ((sn = OBJ_nid2sn(nid)) == NULL))
	{
	    i2t_ASN1_OBJECT(tmp_buf, sizeof(tmp_buf), fname);
	    sn = tmp_buf;
	}
	/* printf("fname is %s len=%d str=%s\n", OBJ_nid2sn(nid), l, str); */

	tuple = PyTuple_New(2);

	PyTuple_SetItem(tuple, 0, PyString_FromString(sn));
	PyTuple_SetItem(tuple, 1, PyString_FromStringAndSize(str, l));

	PyList_SetItem(list, i, tuple);
    }

    return list;
}

/*
 * ADD_METHOD(name) expands to a correct PyMethodDef declaration
 *   {  'name', (PyCFunction)crypto_X509_name, METH_VARARGS }
 * for convenience
 */
#define ADD_METHOD(name)        \
    { #name, (PyCFunction)crypto_X509Name_##name, METH_VARARGS, crypto_X509Name_##name##_doc }
static PyMethodDef crypto_X509Name_methods[] =
{
    ADD_METHOD(add),
    ADD_METHOD(get_hash),
    ADD_METHOD(get_der),
    ADD_METHOD(get_name_components),
    { NULL, NULL }
};
#undef ADD_METHOD

/*
 * Find attribute. An X509Name object has the following attributes:
 * countryName (alias C), stateOrProvince (alias ST), locality (alias L),
 * organization (alias O), organizationalUnit (alias OU), commonName (alias
 * CN) and more...
 *
 * Arguments: self - The X509Name object
 *            name - The attribute name
 * Returns:   A Python object for the attribute, or NULL if something went
 *            wrong
 */
static PyObject *
crypto_X509Name_getattr(crypto_X509NameObj *self, char *name)
{
    int nid;
    PyObject *ret;


    if ((nid = OBJ_txt2nid(name)) != NID_undef)
    {
	ret = get_name_by_nid(self->x509_name, nid);
	if (ret != NULL)
	    return ret;
    }

    return Py_FindMethod(crypto_X509Name_methods, (PyObject *)self, name);
}

/*
 * Set attribute
 *
 * Arguments: self  - The X509Name object
 *            name  - The attribute name
 *            value - The value to set
 */
static int
crypto_X509Name_setattr(crypto_X509NameObj *self, char *name, PyObject *value)
{
    PyObject *strval;
    int nid;

    if ((nid = OBJ_txt2nid(name)) == NID_undef)
    {
        PyErr_SetString(PyExc_AttributeError, "No such attribute");
        return -1;
    }
    
    strval = PyObject_Str(value);
    if (strval == NULL)
        return -1;

    return set_name_by_nid(self->x509_name, nid, strval);
}


/*
 * Compare two X509Name structures.
 *
 * Arguments: n - The first X509Name
 *            m - The second X509Name
 * Returns:   <0 if n < m, 0 if n == m and >0 if n > m
 */
static int
crypto_X509Name_compare(crypto_X509NameObj *n, crypto_X509NameObj *m)
{
    return X509_NAME_cmp(n->x509_name, m->x509_name);
}

/*
 * String representation of an X509Name
 *
 * Arguments: self - The X509Name object
 * Returns:   A string representation of the object
 */
static PyObject *
crypto_X509Name_repr(crypto_X509NameObj *self)
{
    /*
    BIO *bio;
    BUF_MEM *bptr;
    */
    PyObject *ret;
    char *buf;

    /* bio = BIO_new(BIO_s_mem()); */

    /**/
    buf = X509_NAME_oneline(self->x509_name, 0, 0);
    ret = PyString_FromString(buf);
    OPENSSL_free(buf);
    /**/
    
    /*
    X509_NAME_print_ex(bio, self->x509_name, 0, 0);
    BIO_printf(bio, "name=%x", self->x509_name);

    BIO_get_mem_ptr(bio, &bptr);

    printf("got bio len %d max %d\n", bptr->length, bptr->max);

    ret = PyString_FromStringAndSize(bptr->data, bptr->length);

    BIO_set_close(bio, BIO_CLOSE);
    BIO_free(bio);
    */

    return ret;
}

PyTypeObject crypto_X509Name_Type = {
    PyObject_HEAD_INIT(NULL)
    0,
    "X509Name",
    sizeof(crypto_X509NameObj),
    0,
    (destructor)crypto_X509Name_dealloc,
    NULL, /* print */
    (getattrfunc)crypto_X509Name_getattr,
    (setattrfunc)crypto_X509Name_setattr,
    (cmpfunc)crypto_X509Name_compare,
    (reprfunc)crypto_X509Name_repr,
    NULL, /* as_number */
    NULL, /* as_sequence */
    NULL, /* as_mapping */
    NULL  /* hash */
};


/*
 * Initialize the X509Name part of the crypto module
 *
 * Arguments: dict - The crypto module dictionary
 * Returns:   None
 */
int
init_crypto_x509name(PyObject *dict)
{
    crypto_X509Name_Type.ob_type = &PyType_Type;
    Py_INCREF(&crypto_X509Name_Type);
    PyDict_SetItemString(dict, "X509NameType", (PyObject *)&crypto_X509Name_Type);
    return 1;
}
